import TotalTaskCapacity as TTC
import math
import itertools
import SuperPerTask as SPT

class Label():
    def __init__(self, currentNode=1, minTime=0, timeSinceCamp=0, longSinceCamp=False, route=[1]):
        self.currentNode = currentNode
        self.minTime = minTime
        self.timeSinceCamp = timeSinceCamp
        self.longSinceCamp = longSinceCamp
        self.route = route
    def print(self):
        print("currentNode=",self.currentNode, "minTime=", self.minTime, "timeSinceCamp=", self.timeSinceCamp, "longSinceCamp=", self.longSinceCamp , "route=", self.route)

def nVisitsToNode(route, node):
    return route.count(node)

#returnerer en list der verdien representerer om en skill finnes i en av superRessursens subRessurser
#  paramter superResource: et dict med (subRessurs, antall) som (nøkkel, verdi)
#  paramter skillDict: oppslagsverk med (subRessurs, skills) som (nøkkel, verdi)
def generateSuperResourceSkillList(subResources, skillDict, nSkills):
    superResourceSkills = [0 for _ in range(0, nSkills)]
    for subResource in subResources:
        for skillType, subResourceSkill in enumerate(skillDict[subResource]):
            superResourceSkills[skillType] = (subResourceSkill > 0)
    return superResourceSkills

def generateSuperResourceCapacityDict(superResources,skillDict, nSkills):
    superResourceCapacityDict = {}
    for superResource, subResources in superResources.items():
        superResourceCapacityDict[superResource] = [0 for _ in range(nSkills)]
        for subResource in subResources:
            for i in range(nSkills):
                superResourceCapacityDict[superResource][i] += skillDict[subResource][i]
    return superResourceCapacityDict

#maxVisits[0] er et resultat av at python er 0-indeksert og lokasjonene er 1-indeksert
#Dermed blir maxVisits ett element leger enn "nødvendig", men kan da brukes som om den var 1-indeksert.
#maxVisits[1] tilsvarer da max antall mulige besøk på lokasjon 1
def generateSolvableTaskList(nLocations, solvableTasksList, superResources, CapacityRequirement, SRperTask,
                             superResourceCapacityDict, superResource, securityResource, nSkills, tasksAtNode,
                             taskMatrix, superResourceSkills):
    solvableTaskList = [0 for _ in range(0, nLocations+1)]
    for loc in range(1, nLocations+1):
        solvableTaskList[loc] = solvableTasks(superResources, CapacityRequirement, SRperTask, superResourceCapacityDict,
                                              superResource, securityResource, nSkills, tasksAtNode, solvableTasksList,
                                              loc, taskMatrix, superResourceSkills)
    return solvableTaskList

#Returnerer hvor mange ganger en bestemt superressurs kan besøke en bestemt node
def solvableTasks(superResources, CapacityRequirement, SRperTask, superResourceCapacityDict, superResource,
                  securityResource, nSkills, tasksAtNode, solvableTasksList, location, taskMatrix, superResourceSkills):
    tasks = []
    for task in tasksAtNode[location]:
        if task in solvableTasksList:
            for skillType, skill in enumerate(taskMatrix[task-1]):
                if superResourceSkills[skillType] and skill:
                    tasks.append(task)
                    break
    solvableTasksInCombo = []
    for taskWithSkill in tasks:
        canDoTaskCombo = SPT.CanDoTaskInCombination(superResources, taskWithSkill, CapacityRequirement, SRperTask,
                                                        superResourceCapacityDict, superResource, securityResource,
                                                        nSkills)
        if canDoTaskCombo:
            solvableTasksInCombo.append(taskWithSkill)
    return tasks

def generateCanVisit(superResource, node, label, solvableTaskList, taskDeadline, taskRelease, Tmin, Ttask, endTime, Ttravel, belongToTaskLong, belongToTaskSleep, belongToTaskDuplicate):
    minTime = endTime
    canVisit = False
    canDoLong = False
    nSolvableTasks = len(solvableTaskList[node])
    belongToGroup = []
    for dtask in solvableTaskList[node]:
        if belongToTaskDuplicate[dtask-1] in belongToGroup:
            nSolvableTasks = nSolvableTasks - 1
        belongToGroup.append(belongToTaskDuplicate[dtask-1])
    if nSolvableTasks > nVisitsToNode(label.route, node):
        for task in solvableTaskList[node]:
            task = task - 1
            #Om task kan fullføres før deadline
            if taskDeadline[task] >= label.minTime + Ttask[task]*Tmin[task] + Ttravel[superResource-1][label.currentNode-1][node-1]:
                #Om vi kan gjøre task og returnere til BC før sluttid
                if endTime - label.minTime >=  Ttravel[superResource-1][label.currentNode-1][node-1] + Ttask[task]*Tmin[task] + Ttravel[superResource-1][node-1][-1]:
                    if label.longSinceCamp:
                        minTime = min(minTime, max(Ttask[task] * Tmin[task], taskRelease[task] + Ttask[task] * Tmin[task] - label.minTime))
                        canVisit = True
                    # Om rekker tilbake til BC før vi må sove
                    if label.timeSinceCamp + Ttask[task] * Tmin[task] + Ttravel[superResource - 1][node - 1][-1] <= 16 or belongToTaskSleep[task]:
                        minTime = min(minTime, max(Ttask[task] * Tmin[task], taskRelease[task] + Ttask[task] * Tmin[task] - label.minTime))
                        canVisit = True
                    if belongToTaskLong[task]:
                        minTime = min(minTime, max(Ttask[task] * Tmin[task], taskRelease[task] + Ttask[task] * Tmin[task] - label.minTime))
                        canVisit = True
                        canDoLong = True
    return canVisit, minTime, canDoLong

def genSolvableTasksList(nTasks, nSkills, nDuplicates, BelongToTaskDuplicate, superResources, skillDict, CapacityRequirement,
                         Precedence, Connected, DirectStart, nResourcesPerTask, SRperTask, superResourceCapacityDict):
    unsolvableTasksList = []
    subResourceList = []
    for superResource, subResources in superResources.items():
        for subResource in subResources:
            if subResourceList.count(subResource) <= nResourcesPerTask:
                subResourceList.append(subResource)
    for task in range(0, nTasks):
        taskPossibleSub = False
        taskPossibleSuper = False
        # Trekk alle 3 subres kombinasjoner, aggregateSkillList for bare de tre må lages
        subresCombinations = itertools.combinations(subResourceList, nResourcesPerTask)
        tempSet = set(map(tuple, subresCombinations))
        subresCombinationList = map(list, tempSet)
        # for subset in itertools.combinations(subResourceList, nResourcesPerTask):
        for subset in subresCombinationList:
            aggregateSBSkillList = [0 for _ in range(0, nSkills)]
            for SB in subset:
                for skill in range(0, nSkills):
                    aggregateSBSkillList[skill] += skillDict[SB][skill]
            thisCombinationPossible = True
            for skill in range(0, nSkills):
                if CapacityRequirement[task][skill] > aggregateSBSkillList[skill]:
                    thisCombinationPossible = False
                    break
            if thisCombinationPossible:
                taskPossibleSub = True
                break
        # Samme men for 2 og 2 supersres
        superresCombinations = itertools.combinations(superResources.keys(), SRperTask)
        tempSuperSet = set(map(tuple, superresCombinations))
        superresCombinationList = map(list, tempSuperSet)
        for supset in superresCombinationList:
            aggregateSRSkillList = [0 for _ in range(0, nSkills)]
            for SR in supset:
                for skill in range(0, nSkills):
                    aggregateSRSkillList[skill] += superResourceCapacityDict[SR][skill]
            thisCombinationPossible = True
            for skill in range(0, nSkills):
                if CapacityRequirement[task][skill] > aggregateSRSkillList[skill]:
                    thisCombinationPossible = False
                    break
            if thisCombinationPossible:
                taskPossibleSuper = True
                break
        if (not taskPossibleSub) or (not taskPossibleSuper):
            unsolvableTasksList.append(task + 1)

    unsolvableGroupList = []
    for task in unsolvableTasksList:
        if BelongToTaskDuplicate[task-1] not in unsolvableGroupList:
            unsolvableGroupList.append(BelongToTaskDuplicate[task-1])

    for x in range(0, nDuplicates):
        for y in range(0, nDuplicates):
            if (Precedence[x][y] or DirectStart[x][y] or Connected[y][x]) and x+1 in unsolvableGroupList:
                unsolvableTasksList.append(y+1)

    solvableTasksList = []
    for task in range(0,nTasks):
        if BelongToTaskDuplicate[task] not in unsolvableGroupList:
            solvableTasksList.append(task+1)

    return solvableTasksList

def generateLabels(nLocations, endTime, superResource, subResources, skillDict, nSkills,
                   tasksAtLocation, taskMatrix, taskDeadline, taskRelease, Tmin, Ttask,
                   Ttravel, belongToTaskLong, belongToTaskSleep, belongToTaskDuplicate, solvableTasksList,
                   securityResource, value, CapacityRequirement, routeSelection, superResources, SRperTask,
                   superResourceCapacityDict):
    superResourceSkills = generateSuperResourceSkillList(subResources, skillDict, nSkills)
    solvableTaskList = generateSolvableTaskList(nLocations, solvableTasksList, superResources, CapacityRequirement,
                                                SRperTask, superResourceCapacityDict, superResource, securityResource,
                                                nSkills, tasksAtLocation, taskMatrix, superResourceSkills)
    finalLabels = []
    activeLabels = [Label()]
    while activeLabels:
        currentLabel = activeLabels.pop(0)
        ValuePerCapacityNode = []
        NumberCanVisit = 0
        for node in range(3, nLocations + 1):
            canVisit, taskTime, canDoLong = generateCanVisit(superResource, node, currentLabel,
                                                             solvableTaskList, taskDeadline, taskRelease, Tmin,
                                                             Ttask, endTime, Ttravel, belongToTaskLong,
                                                             belongToTaskSleep, belongToTaskDuplicate)
            if canVisit:
                capacityValue = TTC.ValueAtLocation(solvableTaskList, value, node)
                NumberCanVisit += 1
            else:
                capacityValue = 0
            ValuePerCapacityNode.append(capacityValue)
        NumberOfNodes = math.ceil(NumberCanVisit * routeSelection)
        BestNodes = sorted(range(len(ValuePerCapacityNode)), key=lambda x: ValuePerCapacityNode[x])[-NumberOfNodes:]
        BestNodesCorrected = [x + 3 for x in BestNodes]
        for node in range(2, nLocations+1):
            tempLabel = Label()
            if node != currentLabel.currentNode:
                if superResource == securityResource:
                    if node == 2:
                        canVisit = True
                        taskTime = 0
                        canDoLong = False
                    else:
                        canVisit = False
                        taskTime = 0
                        canDoLong = False
                elif node == 2 or node in BestNodesCorrected or node == nLocations:
                    canVisit, taskTime, canDoLong = generateCanVisit(superResource, node, currentLabel,
                                                                     solvableTaskList, taskDeadline, taskRelease, Tmin,
                                                                     Ttask, endTime, Ttravel, belongToTaskLong,
                                                                     belongToTaskSleep, belongToTaskDuplicate)
                else:
                    continue
                if (canVisit and len(currentLabel.route) < 17) or node == nLocations:
                    tempLabel.route = currentLabel.route[:]
                    tempLabel.route.append(node)
                    tempLabel.currentNode = node
                    if node == 2:
                        tempLabel.minTime = currentLabel.minTime + \
                                            Ttravel[superResource - 1][currentLabel.currentNode - 1][node - 1] + taskTime
                        tempLabel.timeSinceCamp = 0
                        tempLabel.longSinceCamp = False
                        activeLabels.append(tempLabel)
                    elif node == nLocations:
                        tempLabel.minTime = currentLabel.minTime + \
                                            Ttravel[superResource - 1][currentLabel.currentNode - 1][
                                                node - 1]
                        tempLabel.timeSinceCamp = 0
                        tempLabel.longSinceCamp = False
                        finalLabels.append(tempLabel)
                    else:
                        tempLabel.minTime = currentLabel.minTime + \
                                            Ttravel[superResource - 1][currentLabel.currentNode - 1][
                                                node - 1] + taskTime
                        tempLabel.timeSinceCamp = currentLabel.timeSinceCamp + tempLabel.minTime - currentLabel.minTime
                        tempLabel.longSinceCamp = (currentLabel.longSinceCamp or canDoLong)
                        activeLabels.append(tempLabel)
    return finalLabels